import re
from typing import Any

p = re.compile(r"(\?)?\.")


def get_obj_attr(obj: object, field: str) -> Any:
    """
    Retrieve nested attribute or dictionary key from an object based on a field string.

    Parameters:
    - obj (object): The object or dictionary from which to extract the attribute or key.
    - field (str): The string specifying the nested attributes or keys, separated by '.'
                   and/or '?.' where '?.' performs safe navigation, returning None if the
                   attribute is not found.

    Returns:
    - Any: The value of the nested attribute or key. Returns None if the attribute is not
           found and safe navigation '?.' is used.

    Examples:
    >>> obj = {'a': {'b': {'c': 1}}}
    >>> get_obj_attr(obj, 'a.b.c')
    1

    >>> class MyClass:
    ...     def __init__(self):
    ...         self.a = {'b': 2}
    >>> obj = MyClass()
    >>> get_obj_attr(obj, 'a.b')
    2

    >>> get_obj_attr(obj, 'a?.c')
    None

    """
    index = 0
    for match in p.finditer(field):
        start, end, s = match.start(), match.end(), match.group()
        if isinstance(obj, dict):
            obj = obj.get(field[index:start])
        else:
            obj = getattr(obj, field[index:start])

        if s == "?." and obj is None:
            return obj
        index = end
    if isinstance(obj, dict):
        obj = obj[field[index:]]
    else:
        obj = getattr(obj, field[index:])
    return obj



from __future__ import annotations

import json
import logging
from io import BytesIO
from typing import Any, Protocol

import xlsxwriter
from xlsxwriter.format import Format
from xlsxwriter.worksheet import Worksheet

logger = logging.getLogger(__name__)


class WorkBook(Protocol):
    def add_format(self, format: dict[str, Any]) -> Format:
        ...

    def worksheets(self) -> list[Worksheet]:
        ...

    def add_worksheet(self, name: str | None = None) -> Worksheet:
        ...


class BytesIOWorkBook:
    def __init__(self) -> None:
        self.io = BytesIO()
        self.workbook = xlsxwriter.Workbook(self.io)

    def add_format(self, format: dict[str, Any]) -> Format:
        return self.workbook.add_format(format)

    def close(self) -> None:
        self.workbook.close()
        self.io.seek(0)

    def read(self) -> bytes:
        data = self.io.read()
        self.io.close()
        return data

    def worksheets(self) -> Worksheet:
        return self.workbook.worksheets()

    def add_worksheet(self, name: str | None = None) -> Worksheet:
        return self.workbook.add_worksheet(name)


class Writer:
    def __init__(
        self,
        workbook: WorkBook,
        worksheet: Worksheet,
        global_format: dict[str, Any] | None = None,
    ) -> None:
        self.workbook = workbook
        self.worksheet = worksheet
        self.global_format = (
            self.workbook.add_format(global_format) if global_format else None
        )
        self.global_format_dict = global_format or {}
        self.formats: dict[str, Any] = {}

    def _calc_format(self, cell_format: Any) -> Any:
        if not cell_format:
            return self.global_format
        elif isinstance(cell_format, dict):
            fmt_key = json.dumps(cell_format, sort_keys=True)
            fmt = self.formats.get(fmt_key)
            if not fmt:
                fmt = self.workbook.add_format(
                    {**self.global_format_dict, **cell_format}
                )
                self.formats[fmt_key] = fmt
            return fmt
        else:
            logger.error(f"cell_format must be dict, got {cell_format}")
            return self.global_format

    def _path_args(self, args: Any) -> Any:
        last_arg = args[-1]
        if isinstance(last_arg, dict):
            args = list(args[:-1]) + [self._calc_format(last_arg)]
        return args

    def write(self, *args: Any) -> None:
        args = self._path_args(args)
        self.worksheet.write(*args)

    def merge_range(self, *args: Any) -> None:
        args = self._path_args(args)
        self.worksheet.merge_range(*args)

    def insert_image(self, *args: Any) -> None:
        self.worksheet.insert_image(*args)



__version__ = "1.1.1"


from .book import Book
from .nodes import Box, Cell, Col, Row, Table
from .sheet import Sheet
from .writer import BytesIOWorkBook

__all__ = ["Sheet", "Book", "Cell", "Box", "Col", "Row", "Table", "BytesIOWorkBook"]



from __future__ import annotations

from io import BytesIO
from typing import Any

from xlsxwriter import Workbook
from xlsxwriter.worksheet import Worksheet

from .nodes import Box, BoxInstance, Col
from .visitors.printer import print_visitor
from .visitors.writer import writer_visitor
from .writer import BytesIOWorkBook, Writer


class Sheet:
    def __init__(
        self,
        root: Box | list[Box],
        start_row: int = 0,
        start_col: int = 0,
        global_format: dict[str, Any] | None = None,
        fast: bool = False,
    ):
        if isinstance(root, list):
            root = Col(children=root)
        BoxInstance(root, start_row, start_col, None)
        self.root = root
        self.global_format = global_format
        self.fast = fast

    @classmethod
    def attach_to_exist_worksheet(
        cls,
        workbook: Workbook,
        worksheet: Worksheet,
        root: Box,
        start_row: int = 0,
        start_col: int = 0,
        global_format: dict[str, Any] | None = None,
    ) -> Writer:
        sheet = cls(root, start_row, start_col)
        writer = Writer(workbook, worksheet, global_format=global_format)
        visitor = writer_visitor(writer, fast=sheet.fast)
        sheet.root.accept(visitor)
        return writer

    def write_to_bytes_io(self) -> BytesIO:
        workbook = BytesIOWorkBook()
        worksheet = workbook.add_worksheet()
        writer = Writer(workbook, worksheet, self.global_format)
        visitor = writer_visitor(writer, fast=self.fast)
        self.root.accept(visitor)
        workbook.close()
        return workbook.io

    def write(self, filename: str) -> None:
        io = self.write_to_bytes_io()
        with open(filename, "wb") as f:
            f.write(io.read())

    def print(self) -> None:
        self.root.accept(print_visitor)

    def to_bytes_io(self) -> BytesIO:
        return self.write_to_bytes_io()



from __future__ import annotations

from io import BytesIO

from .sheet import Sheet
from .visitors.writer import writer_visitor
from .writer import BytesIOWorkBook, Writer


class Book:
    def __init__(self) -> None:
        self.sheets: list[Sheet] = []

    def add_sheet(self, worksheet: Sheet) -> None:
        self.sheets.append(worksheet)

    def write(self, filename: str) -> None:
        data = self.write_to_bytes_io()
        with open(filename, "wb") as f:
            f.write(data.read())
            data.close()

    def write_to_bytes_io(self) -> BytesIO:
        workbook = BytesIOWorkBook()
        for sheet in self.sheets:
            worksheet = workbook.add_worksheet()
            writer = Writer(workbook, worksheet)
            visitor = writer_visitor(writer)
            sheet.root.accept(visitor)

        workbook.close()
        return workbook.io



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['poi', 'poi.visitors']

package_data = \
{'': ['*']}

install_requires = \
['xlsxwriter>=1.1,<2.0']

setup_kwargs = {
    'name': 'poi',
    'version': '1.1.1',
    'description': 'Write Excel XLSX declaratively.',
    'long_description': '# Poi: The Declarative Way to Excel at Excel in Python\n\n![CI](https://github.com/ryanwang520/poi/actions/workflows/tests.yaml/badge.svg)\n\n## Why Poi?\n\nCreating Excel files programmatically has always been a chore. Current libraries offer limited flexibility, especially when you need more than a basic table. That\'s where Poi comes in, offering you a simple, intuitive, yet powerful DSL to make Excel files exactly the way you want them.\n\n\n\n## Installation\n\n```bash\npip install poi\n```\n\n## Quick start\n\n### Create a sheet object and write to a file.\n\n```python\nfrom poi import Sheet, Cell\nsheet = Sheet(\n    root=Cell("hello world")\n)\n\nsheet.write(\'hello.xlsx\')\n```\n\n![hello](https://github.com/baoshishu/poi/raw/master/docs/assets/hello.png)\n\nSee, it\'s pretty simple and clear.\n\n\n### Create a Dynamic Table with Conditional Formatting\n\n\n```python\nfrom typing import NamedTuple\nfrom datetime import datetime\nimport random\n\nfrom poi import Sheet, Table\n\n\nclass Product(NamedTuple):\n    name: str\n    desc: str\n    price: int\n    created_at: datetime\n    img: str\n\n\ndata = [\n    Product(\n        name=f"prod {i}",\n        desc=f"desc {i}",\n        price=random.randint(1, 100),\n        created_at=datetime.now(),\n        img="./docs/assets/product.jpg",\n    )\n    for i in range(5)\n]\ncolumns = [\n    {\n        "type": "image",\n        "attr": "img",\n        "title": "Product Image",\n        "options": {"x_scale": 0.27, "y_scale": 0.25},\n    },\n    ("name", "Name"),\n    ("desc", "Description"),\n    ("price", "Price"),\n    ("created_at", "Create Time"),\n]\nsheet = Sheet(\n    root=Table(\n        data=data,\n        columns=columns,\n        row_height=80,\n        cell_style={\n            "color: red": lambda record, col: col.attr == "price" and record.price > 50\n        },\n        date_format="yyyy-mm-dd",\n        align="center",\n        border=1,\n    )\n)\nsheet.write("table.xlsx")\n```\n\n\n![table](https://github.com/baoshishu/poi/raw/master/docs/assets/table.png)\n\nSee how simple it is to create complex tables? You just wrote a dynamic Excel table with conditional formatting a few lines of code!\n\n\n### Features\n\n* 🎉 Declarative: Create Excel files with a simple, intuitive DSL.\n* 🔥 Fast: Export large Excel files in seconds.\n* 🚀 Flexible Layouts: Create any layout you can imagine with our intuitive Row and Col primitives.\n\n\n### Documentation\n\nFor more details, check our comprehensive [Documentation](https://ryanwang520.github.io/poi/)\n',
    'author': 'Ryan Wang',
    'author_email': 'hwwangwang@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/baoshishu/poi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from functools import singledispatch
from typing import Any

from ..nodes import Cell, Col, Image, Row, Table


@singledispatch
def print_visitor(_: Any) -> None:
    pass


@print_visitor.register
def _(self: Col) -> None:
    print(f"write Col at {self.row}:{self.col}")
    for child in self.children:
        print_visitor(child)


@print_visitor.register  # type: ignore
def _(self: Row):
    print(f"write Row at {self.row}:{self.col}")
    for child in self.children:
        print_visitor(child)


@print_visitor.register  # type: ignore
def _(self: Table):  # type: ignore
    print(f"write Table at {self.row}:{self.col}")


@print_visitor.register  # type: ignore
def _(self: Cell):
    print(
        f"write Cell {self.value} at {self.row}:{self.col} "
        f"rowspan {self.rowspan} colspan {self.colspan}"
    )


@print_visitor.register  # type: ignore
def _(self: Image):
    print(f"insert image {self.filename} at {self.row}:{self.col}")



from __future__ import annotations

import logging
from collections import abc
from typing import (
    Any,
    Callable,
    Collection,
    Generic,
    Iterable,
    NamedTuple,
    TypeVar,
)

try:
    from typing_extensions import Literal
except ImportError:
    from typing import Literal  # type: ignore

logger = logging.getLogger("poi")
logger.addHandler(logging.NullHandler())

Direction = Literal["HORIZONTAL", "VERTICAL"]


class BoxInstance:
    def __init__(
        self, box: Box, row: int, col: int, parent: BoxInstance | None
    ) -> None:
        self.box = box
        self.row = row
        self.col = col
        self.parent = parent
        box.instance = self
        self._setup_children()

    def _setup_children(self) -> None:
        children = self.box.children
        if not children:
            return
        children = [
            Cell(child) if isinstance(child, str) else child for child in children
        ]

        self.children = []

        current_row = self.row
        current_col = self.col
        for child in children:
            child.styles = {**child.styles, **self.box.styles}
            self.box.add_child_span(
                child, neighbours=[c for c in children if c is not child]
            )
            child_node, current_row, current_col = self.box.layout_child_element(
                child, current_row, current_col
            )
            self.children.append(child_node)


_NotDetermined = object()


Visitor = Callable[["Box"], None]


class Box:
    instance: BoxInstance
    parent: Box | None

    def accept(self, visitor: Visitor) -> None:
        visitor(self)

    @property
    def row(self) -> int:
        assert self.instance
        return self.instance.row

    @property
    def col(self) -> int:
        assert self.instance
        return self.instance.col

    @property
    def direction(self) -> Direction:
        if not self.parent:
            raise ValueError("root node cannot have direction")
        if isinstance(self.parent, Row):
            return "HORIZONTAL"
        if isinstance(self.parent, Col):
            return "VERTICAL"
        raise ValueError(f"invalid parent {self.parent}")

    @property
    def is_horizontal(self) -> bool:
        return self.direction == "HORIZONTAL"

    @property
    def is_vertical(self) -> bool:
        return self.direction == "VERTICAL"

    children: list[Box]

    def __init__(
        self,
        children: Iterable[Box | None] | Box | None = None,
        rowspan: int | None = None,
        colspan: int | None = None,
        offset: int = 0,
        grow: bool = False,
        **kwargs: Any,
    ) -> None:
        self.parent = None
        self.rowspan = rowspan
        self.colspan = colspan
        self.offset = offset
        self.grow = grow

        def flatten(items: Any) -> Iterable[Any]:
            """Yield items from any nested iterable"""
            for x in items:
                if isinstance(x, abc.Iterable) and not isinstance(x, (str, bytes)):
                    yield from flatten(x)
                else:
                    yield x

        if isinstance(children, Box):
            children = [children]
        else:
            children = [child for child in flatten(children or []) if child is not None]
        self.children = children  # type: ignore
        for child in self.children:
            child.parent = self
        self.styles = kwargs
        self.instance = None  # type: ignore

    def add_child_span(self, child: Box, neighbours: list[Box]) -> None:
        pass

    @property
    def cell_format(self) -> dict[str, Any]:
        return self.styles or {}

    def layout_child_element(
        self, child: Box, current_row: int, current_col: int
    ) -> Any:
        raise ValueError(f"not support type for {self}")

    def __repr__(self) -> str:
        if not self.instance:
            return f"""
        Unbound box {self.__class__.__name__}
        rowspan {self.rowspan}
        colspan {self.colspan}
            """
        return f"""
        {self.__class__.__name__}
        row {self.row}
        col {self.col}
        rowspan {self.rowspan}
        colspan {self.colspan}
        """

    @property
    def cols(self) -> int:
        raise NotImplementedError

    @property
    def rows(self) -> int:
        raise NotImplementedError

    def assert_children_bound(self) -> None:
        assert all(child.instance for child in self.children)

    def calculate_column_span(self, raises: bool = True) -> Any:
        if self.colspan:
            return self.colspan + self.offset
        if self.grow:
            if raises:
                raise ValueError(f"{self} cannot grow as the col have to be determined")
            return _NotDetermined
        max_col = 1
        for child in self.children:
            if child.colspan:
                if max_col < child.colspan:
                    max_col = child.colspan
            else:
                col = child.calculate_column_span(raises=raises)
                if col is not _NotDetermined and col > max_col:
                    max_col = col
        return max_col + self.offset

    def calculate_row_span(self, raises: bool = True) -> Any:
        if self.rowspan:
            return self.rowspan + self.offset
        if self.grow:
            if raises:
                raise ValueError(f"{self} cannot grow as the row have to be determined")
            return _NotDetermined
        max_row = 1
        for child in self.children:
            if child.rowspan:
                if max_row < child.rowspan:
                    max_row = child.rowspan
            else:
                row = child.calculate_row_span(raises=raises)
                if row is not _NotDetermined and row > max_row:
                    max_row = row
        return max_row + self.offset


class Row(Box):
    colspan: int

    def layout_child_element(
        self, child: Box, current_row: int, current_col: int
    ) -> tuple[BoxInstance, int, int]:
        child_node = BoxInstance(
            child, current_row, current_col + child.offset, self.instance
        )
        current_col += child.cols
        return child_node, current_row, current_col

    def add_child_span(self, child: Box, neighbours: list[Box]) -> None:
        if not child.rowspan:
            child.rowspan = self.rowspan
        if child.grow:
            assert all(
                not c.grow for c in neighbours
            ), "only one col in a row can have grow attr"
            if not self.colspan:
                if self.instance.parent:
                    parent = self.instance.parent.box
                else:
                    raise ValueError(f"{child} width is not determinable")

                neighbor_with_cols = [child for child in parent.children if child]
                if neighbor_with_cols:
                    neighbour_cols = [
                        n.calculate_column_span(raises=False)
                        for n in neighbor_with_cols
                    ]
                    valid_cols = [
                        col for col in neighbour_cols if col is not _NotDetermined
                    ]
                    if valid_cols:
                        self.colspan = max(valid_cols)
                    else:
                        raise ValueError(f"{child} width is not determinable")
                else:
                    raise ValueError(f"{child} width is not determinable")
            child.colspan = (
                self.colspan
                - child.offset
                - sum(c.calculate_column_span() for c in neighbours)
            )

    @property
    def cols(self) -> int:
        """
        cols and row can only be accessed if all chilren are bound to instance
        """
        offset = self.offset if self.is_horizontal else 0
        if self.colspan:
            return self.colspan + offset
        self.assert_children_bound()
        return sum(child.cols for child in self.children) + offset

    @property
    def rows(self) -> int:
        offset = self.offset if self.is_vertical else 0
        if self.rowspan:
            return self.rowspan + offset
        self.assert_children_bound()
        return max(child.rows for child in self.children) + offset


class Col(Box):
    rowspan: int

    def add_child_span(self, child: Box, neighbours: list[Box]) -> None:
        if not child.colspan:
            child.colspan = self.colspan
        if child.grow:
            assert all(
                not c.grow for c in neighbours
            ), "only one row in a col can have grow attr"
            if not self.rowspan:
                if self.instance.parent:
                    parent = self.instance.parent.box
                else:
                    raise ValueError(f"{child} height is not determinable")
                neighbor_with_rows = [child for child in parent.children if child]
                if neighbor_with_rows:
                    neighbour_rows = [
                        n.calculate_row_span(raises=False) for n in neighbor_with_rows
                    ]
                    valid_rows = [
                        row for row in neighbour_rows if row is not _NotDetermined
                    ]
                    if valid_rows:
                        self.rowspan = max(valid_rows)
                    else:
                        raise ValueError(f"{child} height is not determinable")
                else:
                    raise ValueError(f"{child} height is not determinable")
            child.rowspan = (
                self.rowspan
                - child.offset
                - sum(c.calculate_row_span() for c in neighbours)
            )

    def layout_child_element(
        self, child: Box, current_row: int, current_col: int
    ) -> tuple[BoxInstance, int, int]:
        child_node = BoxInstance(
            child, current_row + child.offset, current_col, self.instance
        )
        current_row += child.rows
        return child_node, current_row, current_col

    @property
    def cols(self) -> int:
        if self.colspan:
            offset = self.offset if self.is_horizontal else 0
            return self.colspan + offset
        self.assert_children_bound()
        return max(child.cols for child in self.children)

    @property
    def rows(self) -> int:
        if self.rowspan:
            offset = self.offset if self.is_vertical else 0
            return self.rowspan + offset
        self.assert_children_bound()
        return sum(child.rows for child in self.children)


class PrimitiveBox(Box):
    @property
    def cols(self) -> int:
        offset = self.offset if self.is_horizontal else 0
        return (self.colspan or 1) + offset

    @property
    def rows(self) -> int:
        offset = self.offset if self.is_vertical else 0
        return (self.rowspan or 1) + offset


class Cell(PrimitiveBox):
    def __init__(
        self,
        value: str,
        *args: Any,
        width: int | None = None,
        height: int | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.value = value
        self.height = height
        self.width = width


class Image(PrimitiveBox):
    def __init__(self, filename: str, *args: Any, **kwargs: Any) -> None:
        options = kwargs.pop("options", None)
        super().__init__(*args, **kwargs)
        self.filename = filename
        self.options = options


class Column(NamedTuple):
    title: str
    attr: str | None = None
    render: Callable[..., Any] | None = None
    width: int | None = None
    type: Literal["image", "text"] = "text"
    options: dict[str, Any] | None = None
    format: dict[str, Any] | None = None


T = TypeVar("T")


class Table(Box, Generic[T]):
    columns: list[Column]

    def __init__(
        self,
        data: Collection[T],
        columns: Collection[dict[str, Any] | tuple[str, str]],
        col_width: int | None = None,
        row_height: Callable[..., Any] | int | None = None,
        border: int | None = None,
        cell_style: (
            dict[str, Callable[[T, Column], bool] | Callable[[T], bool]] | str | None
        ) = None,
        datetime_format: str | None = None,
        date_format: str | None = None,
        time_format: str | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, border=border or 1, **kwargs)
        self.data = data
        self.col_width = col_width or 15
        self.row_height = row_height

        self.cell_style = cell_style or {}
        self.date_format = date_format
        self.datetime_format = datetime_format
        self.time_format = time_format
        self.columns = []
        for col in columns:
            assert isinstance(col, (tuple, dict))
            if isinstance(col, tuple):
                item = Column(attr=col[0], title=col[1])
            else:
                item = Column(
                    attr=col.get("attr"),
                    title=col["title"],
                    type=col.get("type"),  # type: ignore
                    options=col.get("options"),
                    render=col.get("render"),
                    width=col.get("width"),
                    format=col.get("format"),
                )
            self.columns.append(item)

        self.rowspan = len(self.data) + 1
        self.colspan = len(self.columns)

    @property
    def rows(self) -> int:
        offset = self.offset if self.is_vertical else 0
        return self.rowspan + offset  # type: ignore

    @property
    def cols(
        self,
    ) -> int:
        offset = self.offset if self.is_horizontal else 0
        assert self.colspan
        return self.colspan + offset



import datetime
import re
from functools import singledispatch
from inspect import signature
from typing import Any, Callable, Dict

from ..nodes import Cell, Col, Image, Row, Table
from ..utils import get_obj_attr
from ..writer import Writer


def call_by_sig(fn: Callable[..., Any], *args: Any) -> Any:
    sig = signature(fn)
    arg_length = len(sig.parameters)
    return fn(*args[:arg_length])


def writer_visitor(writer: Writer, fast=False) -> Any:
    EMPTY_VALUES = (None, "")

    def should_write(value: object) -> bool:
        if not fast:
            return True
        return value not in EMPTY_VALUES

    @singledispatch
    def visitor(_: Any) -> None:
        pass

    @visitor.register
    def _(self: Row) -> None:
        for child in self.children:
            visitor(child)

    @visitor.register
    def _(self: Col) -> None:
        for child in self.children:
            visitor(child)

    @visitor.register
    def _(self: Table) -> None:  # type: ignore
        row, col = self.row, self.col
        for i in range(len(self.data) + 1):
            height = None
            if self.row_height and i >= 1:
                if isinstance(self.row_height, int):
                    height = self.row_height
                else:
                    height = call_by_sig(self.row_height, self.data[i - 1], i - 1)  # type: ignore
            if height:
                writer.worksheet.set_row(self.row + i, height)
        for i, column in enumerate(self.columns):
            width = column.width or self.col_width
            if width:
                writer.worksheet.set_column(self.col + i, self.col + i, width)
            if should_write(column.title):
                writer.write(row, col + i, column.title, self.cell_format)

        for i, item in enumerate(self.data):

            def format_from_style(style_css: str) -> Dict[str, Any]:
                rv = {}
                for style in style_css.split(";"):
                    style = style.strip()
                    if style == "":
                        continue
                    k, v = re.split(r"\s*:\s*", style)
                    rv[k] = v
                return rv

            for j, column in enumerate(self.columns):
                fmt = {}
                if isinstance(self.cell_style, str):
                    fmt.update(format_from_style(self.cell_style))

                else:
                    for styles, condition in self.cell_style.items():
                        if call_by_sig(condition, item, column):
                            fmt.update(format_from_style(styles))
                if column.attr:
                    val = get_obj_attr(item, column.attr)
                else:
                    assert column.render
                    val = call_by_sig(column.render, item, column)
                if isinstance(val, datetime.datetime):
                    fmt["num_format"] = self.datetime_format or "yyyy-mm-dd hh:mm:ss"
                if isinstance(val, datetime.date):
                    fmt["num_format"] = self.date_format or "yyyy-mm-dd"
                if isinstance(val, datetime.time):
                    fmt["num_format"] = self.time_format or "hh:mm:ss"

                if column.format:
                    fmt.update(column.format)

                if column.type == "image":
                    writer.insert_image(row + i + 1, col + j, val, column.options)
                else:
                    if should_write(val):
                        writer.write(
                            row + i + 1, col + j, val, {**self.cell_format, **fmt}
                        )

    @visitor.register
    def _(self: Image) -> None:
        writer.insert_image(self.row, self.col, self.filename, self.options)

    @visitor.register
    def _(self: Cell) -> None:
        colspan = self.colspan or 1
        rowspan = self.rowspan or 1
        if self.width:
            writer.worksheet.set_column(self.col, self.col + colspan - 1, self.width)
        if rowspan == 1 and self.height:
            writer.worksheet.set_row(self.row, self.height)
        if not should_write(self.value):
            return
        if colspan == 1 and rowspan == 1:
            writer.write(self.row, self.col, self.value, self.cell_format)
        else:
            writer.merge_range(
                self.row,
                self.col,
                self.row + rowspan - 1,
                self.col + colspan - 1,
                self.value,
                self.cell_format,
            )

    return visitor
